import { type Readable } from 'svelte/store';
import { type z } from 'zod';
import { ZodFieldStore } from './ZodFieldStore.js';
/**
 * Settings for ZodFormStore
 */
interface ICreateFormOptions<T> {
    /**
     * The initial value of fields in the form.
     */
    initialValue?: Partial<T>;
    /**
     * Async callback to handle submmition of the form.
     * Should return nothing, or an `string` contain error message
     */
    onSubmit?: (v: T) => Promise<void | string> | string | void;
    /**
     * Auto trigger submit when any data changed, after the delay in `ms`.
     * Passing `0` or `undefined` to disabled.
     *
     * @default undefined
     */
    autoSubmitAfter?: number;
    /**
     * Debounce the value update, in `ms`.
     * Passing falsy value (`0` or `undefined`) to disabled.
     *
     * @default undefined
     */
    debounceDelay?: number;
    /** Print debug messages */
    debug?: boolean;
}
/**
 * Instance that hold all our form's state, as Svelte's Store
 */
export declare class ZodFormStore<A extends z.ZodRawShape = z.ZodRawShape, O = A> {
    /** Form's data. Will be passed to onSubmit handler */
    readonly model: Readable<O>;
    /** Form settings. Should not be update */
    readonly options: ICreateFormOptions<O>;
    /**
     * Generated fields's functions
     */
    fields: {
        [K in Extract<keyof O, string>]: Pick<ZodFieldStore<K, A, O>, 'updateValue' | 'setValue' | 'handleChange' | 'handleBlur' | 'reset' | 'setError' | 'setTouched'>;
    };
    /** Generated fields's stores */
    stores: {
        [K in Extract<keyof O, string> as `${K}_value`]: ZodFieldStore<K, A, O>['value'];
    } & {
        [K in Extract<keyof O, string> as `${K}_touched`]: ZodFieldStore<K, A, O>['touched'];
    } & {
        [K in Extract<keyof O, string> as `${K}_dirty`]: ZodFieldStore<K, A, O>['dirty'];
    } & {
        [K in Extract<keyof O, string> as `${K}_error`]: ZodFieldStore<K, A, O>['error'];
    } & {
        [K in Extract<keyof O, string> as `${K}_valid`]: ZodFieldStore<K, A, O>['valid'];
    };
    /**
     * Function to start parsing, validating and submit the form's data.
     *
     * Should assign to form's submit event or submit button.
     *
     * ```svelte
     * <form on:submit={form.triggerSubmit}>
     *  ...
     * </form>
     * ```
     */
    triggerSubmit: () => Promise<void>;
    /**
     * Function to reset the form to original state.
     *
     * Should assign to form's `reset` event, reset button, or call directly.
     *
     * ```svelte
     * <form on:reset={form.reset}>
     *  ...
     * </form>
     * ```
     */
    reset: () => void;
    /** Indicate if the form is being submitted (`onSubmit` handler is resolving).  */
    submitting: Readable<boolean>;
    /**
     * Array of string contains all error messages
     * (including fields's errors and error return from `onSubmit` handler).
     */
    errors: Readable<string[]>;
    /** Error message returned from `onSubmit` handler, or custom validation message. */
    error: Readable<string>;
    /** Indicate if the form is edited or submitted. */
    dirty: Readable<boolean>;
    /** Indicate if the form is valid. */
    valid: Readable<boolean>;
    constructor(
    /** Zod's schema for data in the form. Should be a `ZodObject`. */
    schema: z.ZodObject<A, z.UnknownKeysParam, z.ZodTypeAny, O> | z.ZodEffects<z.ZodObject<A, z.UnknownKeysParam, z.ZodTypeAny, O>>, 
    /** Additional configuration */
    options?: ICreateFormOptions<O>);
    /** Set the form's error message manually */
    setError(errorMessage: string, path: (string | number)[]): void;
}
export {};
