import { derived, writable, get } from 'svelte/store';
import { ZodError, ZodEffects } from 'zod';

function zip(...arrays) {
  if (!arrays || !arrays.length)
    return [];
  return new Array(Math.max(...arrays.map(({ length }) => length))).fill([]).map((_, idx) => arrays.map((array) => array[idx]));
}

const debounce = ({ delay }, func) => {
  let timer = void 0;
  let active = true;
  const debounced = (...args) => {
    if (active) {
      clearTimeout(timer);
      timer = setTimeout(() => {
        active && func(...args);
        timer = void 0;
      }, delay);
    } else {
      func(...args);
    }
  };
  debounced.isPending = () => {
    return timer !== void 0;
  };
  debounced.cancel = () => {
    active = false;
  };
  debounced.flush = (...args) => func(...args);
  return debounced;
};

const pick = (obj, keys2) => {
  if (!obj)
    return {};
  return keys2.reduce((acc, key) => {
    if (Object.prototype.hasOwnProperty.call(obj, key))
      acc[key] = obj[key];
    return acc;
  }, {});
};

const trurthly = (z) => !!z;
const toReadable = (w) => derived(w, a => a);
const getErrorMessage = (e) => {
    if (e instanceof ZodError)
        return e.issues
            .filter(a => a.path.length === 0) // Only take error from root
            .map(a => a.message)
            .join();
    if (e instanceof Error)
        return e.message;
    if (typeof e === 'string')
        return e;
    if (typeof e === 'undefined')
        return '';
    if (e === null)
        return '';
    return JSON.stringify(e);
};

class ZodFieldStore {
    /**
     * Field name
     */
    name;
    /**
     * Readable store holding field's value
     */
    value;
    /**
     * The field have been touched or not
     */
    touched;
    /**
     * The field value been changed or not
     */
    dirty;
    /**
     * The field validation error, if any
     */
    error;
    /**
     * The field value is valid or not
     */
    valid;
    /**
     * Callback to update field's value
     */
    updateValue;
    /**
     * Function to set field's value
     */
    setValue;
    /**
     * Callback to update field's value
     */
    handleChange;
    /**
     * Callback to mark field as touched
     */
    handleBlur;
    /**
     * Reset field to original state
     */
    reset;
    /**
     * Set custom field error
     */
    setError;
    /**
     * Update touched state
     */
    setTouched;
    constructor(parentSchema, name, initialValue, ms = 0) {
        if (!(name in parentSchema.shape))
            throw new Error('Invalid field name');
        // Stores
        const touched = writable(false);
        const dirty = writable(false);
        const error = writable('');
        const value = writable();
        const setError = ms > 0 ? debounce({ delay: ms }, error.set) : error.set;
        const setTouched = ms > 0 ? debounce({ delay: ms }, touched.set) : touched.set;
        const setDirty = ms > 0 ? debounce({ delay: ms }, dirty.set) : dirty.set;
        const schema = parentSchema.shape[name];
        const handleError = (e) => setError(getErrorMessage(e));
        try {
            initialValue = schema.parse(initialValue);
        }
        catch (e) {
            handleError(e);
        }
        const _resetValue = () => value.set(initialValue);
        _resetValue();
        const _setValue = ms > 0 ? debounce({ delay: ms }, value.set) : value.set;
        const _updateValue = ms > 0 ? debounce({ delay: ms }, value.update) : value.update;
        const handleChange = (e) => {
            setError('');
            let nextVal = e;
            if (e instanceof Event)
                if (e instanceof CustomEvent)
                    if (typeof e.detail === 'object' && e.detail !== null && 'value' in e.detail)
                        nextVal = e.detail.value;
                    else
                        nextVal = e.detail;
                else if (e.currentTarget instanceof HTMLInputElement)
                    if (e.currentTarget.type === 'checkbox')
                        nextVal = e.currentTarget.checked;
                    else
                        nextVal = e.currentTarget.value;
                else if (e.currentTarget instanceof HTMLSelectElement ||
                    e.currentTarget instanceof HTMLTextAreaElement)
                    nextVal = e.currentTarget.value;
            try {
                nextVal = schema.parse(nextVal);
            }
            catch (e) {
                handleError(e);
            }
            _setValue(nextVal);
            setTouched(true);
            setDirty(true);
        };
        const handleSetValue = (v) => {
            setError('');
            if (schema)
                try {
                    _setValue(schema.parse(v));
                }
                catch (e) {
                    handleError(e);
                }
            else
                _setValue(v);
            setTouched(true);
            setDirty(true);
        };
        const handleUpdateValue = (updater) => {
            const safeUpdater = (v) => {
                setError('');
                const updatedVal = updater(v);
                if (schema)
                    try {
                        return schema.parse(updatedVal);
                    }
                    catch (e) {
                        handleError(e);
                    }
                return updatedVal;
            };
            _updateValue(safeUpdater);
            setTouched(true);
            setDirty(true);
        };
        const handleBlur = () => setTouched(true);
        const handleReset = () => {
            touched.set(false);
            dirty.set(false);
            error.set('');
            _resetValue();
        };
        this.name = name;
        // Stores
        this.value = toReadable(value);
        this.touched = toReadable(touched);
        this.dirty = toReadable(dirty);
        this.error = toReadable(error);
        this.valid = derived(error, (e) => !e);
        // Handlers
        this.handleChange = handleChange;
        this.updateValue = handleUpdateValue;
        this.setValue = handleSetValue;
        this.handleBlur = handleBlur;
        this.reset = handleReset;
        this.setError = setError;
        this.setTouched = setTouched;
    }
}

/**
 * Instance that hold all our form's state, as Svelte's Store
 */
class ZodFormStore {
    /** Form's data. Will be passed to onSubmit handler */
    model;
    /** Form settings. Should not be update */
    options;
    /**
     * Generated fields's functions
     */
    fields;
    /** Generated fields's stores */
    // prettier-ignore
    stores;
    /**
     * Function to start parsing, validating and submit the form's data.
     *
     * Should assign to form's submit event or submit button.
     *
     * ```svelte
     * <form on:submit={form.triggerSubmit}>
     *  ...
     * </form>
     * ```
     */
    triggerSubmit;
    /**
     * Function to reset the form to original state.
     *
     * Should assign to form's `reset` event, reset button, or call directly.
     *
     * ```svelte
     * <form on:reset={form.reset}>
     *  ...
     * </form>
     * ```
     */
    reset;
    /** Indicate if the form is being submitted (`onSubmit` handler is resolving).  */
    submitting;
    /**
     * Array of string contains all error messages
     * (including fields's errors and error return from `onSubmit` handler).
     */
    errors;
    /** Error message returned from `onSubmit` handler, or custom validation message. */
    error;
    /** Indicate if the form is edited or submitted. */
    dirty;
    /** Indicate if the form is valid. */
    valid;
    constructor(
    /** Zod's schema for data in the form. Should be a `ZodObject`. */
    schema, 
    /** Additional configuration */
    options) {
        this.options = options || {};
        // Form stores
        const submitting = writable(false);
        const dirty = writable(false);
        const formError = writable();
        // Fields stores
        const fieldNames = Object.keys(schema instanceof ZodEffects ? schema.innerType().shape : schema.shape);
        const fieldStores = fieldNames.map((name) => new ZodFieldStore(schema instanceof ZodEffects ? schema.innerType() : schema, name, this.options.initialValue?.[name], this.options.debounceDelay));
        const model = derived(fieldStores.map((z) => z.value), (z) => Object.fromEntries(zip(fieldNames, z)));
        const fieldErrors = derived(fieldStores.map((z) => z.error), (e) => e.filter(trurthly));
        const triggerSubmit = async () => {
            dirty.set(true);
            fieldStores.forEach((f) => f.setTouched(true));
            if (!this.options.onSubmit)
                return;
            const _fieldErrors = get(fieldErrors);
            if (_fieldErrors.length > 0) {
                if (this.options.debug)
                    console.debug('❗️ Form not valid', _fieldErrors);
                return;
            }
            formError.set('');
            submitting.set(true);
            try {
                const values = schema.parse(get(model));
                const result = await this.options.onSubmit(values);
                if (result)
                    formError.set(result);
            }
            catch (e) {
                if (this.options.debug)
                    console.debug(`🚫 Form submit error`, e);
                // Set Error back to field
                if (e instanceof ZodError)
                    for (const issue of e.issues)
                        this.setError(issue.message, issue.path);
                formError.set(getErrorMessage(e));
            }
            finally {
                submitting.set(false);
            }
        };
        function handleReset() {
            dirty.set(false);
            submitting.set(false);
            formError.set('');
            fieldStores.forEach((f) => f.reset());
        }
        if (this.options.autoSubmitAfter && this.options.autoSubmitAfter > 0) {
            const debouncedTriggerSubmit = debounce({ delay: this.options.autoSubmitAfter }, () => {
                if (this.options.debug)
                    console.debug('⏱️ Auto submitting...');
                triggerSubmit();
            });
            let prevModel = null;
            model.subscribe((nextModel) => {
                if (prevModel === null) {
                    prevModel = nextModel;
                    return;
                }
                if (prevModel === nextModel)
                    return;
                prevModel = nextModel;
                debouncedTriggerSubmit();
            });
        }
        const handlerFuncNames = [
            'updateValue',
            'setValue',
            'handleChange',
            'handleBlur',
            'reset',
            'setError',
            'setTouched',
        ];
        this.fields = Object.fromEntries(fieldStores.map((z) => [z.name, pick(z, handlerFuncNames)]));
        this.stores = Object.fromEntries(fieldStores.flatMap((fieldStore) => Object.keys(fieldStore)
            .filter((key) => key !== 'name' && !handlerFuncNames.includes(key))
            .map((prop) => [
            `${String(fieldStore.name)}_${prop}`,
            fieldStore[prop],
        ])));
        this.triggerSubmit = triggerSubmit;
        this.reset = handleReset;
        this.submitting = toReadable(submitting);
        this.error = toReadable(formError);
        this.dirty = derived([dirty, ...fieldStores.map((a) => a.dirty)], (a) => a.some((b) => b));
        this.errors = derived([fieldErrors, formError], (errors) => errors.flatMap((z) => z).filter(trurthly));
        this.model = model;
        this.valid = derived(this.errors, (e) => !e.length);
    }
    /** Set the form's error message manually */
    setError(errorMessage, path) {
        const [currentPath] = path;
        const field = this.fields[currentPath];
        if (!field)
            return;
        field.setError(errorMessage);
    }
}

export { ZodFormStore };
